using System;
using System.Collections;

namespace Flibs {
    public class Engine : IChildCreator {
        private class FlibFitnessComparer : IComparer {
            public int Compare(object x, object y) {
                return -1*((Flib) x).Fitness.CompareTo(((Flib) y).Fitness);
            }
        }

        private static Random _random = new Random((int) DateTime.Now.Ticks);
        private ExecuteData _executeData;
        private IList _generation;
        private double _fitness = 0;
        private int _stepCount = 0;

        private double _bestFitness = 0;
        private double _worstFitnes = 0;
        private double _averageFitness = 0;
        private Flib _bestFoundFlib;

        public void Execute() {
            while (ExecuteStep()) ;
        }

        public bool ExecuteStep() {
            if (_stepCount >= _executeData.GenerationCount) return false;

            _executeData.SignalSource.Nulling();
            NullingFlibs();

            for (int j = 0; j < _executeData.CheckCount; j++) {
                DoStep();
            }

            Flib bestFlib = GetBestFlib(_generation);
            if (_bestFoundFlib == null || _bestFoundFlib.Fitness < bestFlib.Fitness)
                _bestFoundFlib = bestFlib.Clone();

            _bestFitness = bestFlib.Fitness;
            _worstFitnes = ((Flib) _generation[GetWorstFlibIndex(_generation)]).Fitness;
            _averageFitness = CalculateAverageFitness();

            if (_fitness < _bestFitness) 
                _fitness = _bestFitness;

            _generation = _executeData.GenerationCreator.CreateGenearation(_random, _generation, this);

            _stepCount++;
            return true;
        }

        private double CalculateAverageFitness() {
            double summ = 0;
            foreach (Flib flib in _generation) {
                summ += flib.Fitness;
            }

            return summ/_generation.Count;
        }

        private void NullingFlibs() {
            foreach (Flib flib in _generation) {
                flib.Nulling();
            }
        }

        public Engine(ExecuteData data) {
            _executeData = data;
            CreateRandomGeneration();
        }

        public void DoStep() {
            _executeData.SignalSource.DoStep();

            foreach (Flib flib in _generation) {
                flib.Step(_executeData.SignalSource.Input, _executeData.SignalSource.InputNext);
            }
        }

        public void CreateRandomGeneration() {
            _generation = new ArrayList(_executeData.GenerationSize);

            while (_generation.Count < _executeData.GenerationSize) {
                Flib flib = CreateRandomFlib(_random, _executeData.StateCount);
                new FlibRestorer().Restore(_random, flib);
                _generation.Add(flib);
            }
        }

        public static Flib CreateRandomFlib(Random _random, int stateCount) {
            Flib flib = new Flib(stateCount);

            for (int i = 0; i < flib.States.Length; i++) {
                flib.States[i] = CreateRandomState(_random, stateCount);
            }

            return flib;
        }

        public static State CreateRandomState(Random random, int stateCount) {
            State state = new State();

            for (int i = 0; i < Branch.TARGET_COUNT; i++) {
                state.Branches[i] = CreateRandomBranch(random, stateCount);
            }

            return state;
        }

        public static Branch CreateRandomBranch(Random random, int stateCount) {
            return new Branch(random.Next(stateCount), random.Next(Branch.TARGET_COUNT));
        }

        public Flib GetBestFlib(ICollection generation) {
            Flib result = null;
            double fitness = int.MinValue;

            foreach (Flib flib in generation) {
                if (fitness < flib.Fitness) {
                    fitness = flib.Fitness;
                    result = flib;
                }
            }

            return result;
        }

        private Flib this[int index] {
            get { return (Flib) _generation[index]; }
            set { _generation[index] = value; }
        }

        public int GetWorstFlibIndex(IList generation) {
            int result = 0;

            for (int i = 1; i < generation.Count; i++) {
                if (this[result].Fitness > this[i].Fitness)
                    result = i;
            }

            return result;
        }

        public Flib CreateChild(Flib firstParent, Flib secondParent) {
            Flib child = _executeData.Crossover.CreateChild(_random, firstParent, secondParent);

            _executeData.Mutation.Mutate(_random, child);

            if (_executeData.IsRestoreFlibs)
                new FlibRestorer().Restore(_random, child);

            return child;
        }

        public ICollection GetBestFlibs(ICollection flibs, int count) {
            if (count == 0)
                return new ArrayList();

            ArrayList sortedList = new ArrayList();
            foreach (Flib flib in flibs) {
                if (sortedList.Count == count && ((Flib) sortedList[count - 1]).Fitness < flib.Fitness)
                    sortedList.RemoveAt(count - 1);

                if (sortedList.Count < count) {
                    sortedList.Add(flib);
                    sortedList.Sort(new FlibFitnessComparer());
                }
            }

            return sortedList;
        }

        public double BestFitness {
            get { return _bestFitness; }
        }

        public double WorstFitnes {
            get { return _worstFitnes; }
        }

        public double AverageFitness {
            get { return _averageFitness; }
        }

        public Flib BestFoundFlib {
            get { return _bestFoundFlib; }
        }
    }
}